from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

import jijzept.response.base
import jijzept.response.base as base
import jijzept.response.dimodresopnse
import jijzept.response.dimodresopnse as dimodresopnse

from jijzept.response.base import APIStatus, BaseResponse
from jijzept.response.dimodresopnse import DimodResponse

__all__ = [
    "APIStatus",
    "BaseResponse",
    "DimodResponse",
]



from __future__ import annotations
from dataclasses import dataclass, asdict
from numbers import Number
from typing import Optional, Union
import logging

import numpy as np

from jijzept.response import DimodResponse
from jijmodeling import Problem
from jijzept.client import JijZeptClient
from jijzept.entity.schema import SolverType
from jijzept.sampler.jijmodel_post import JijModelingInterface


@dataclass
class JijFixstarsAmplifyParameters:
    """Manage Parameters for using Fixstars Amplify
    Attributes:
        amplify_timeout (int): Set the timeout in milliseconds. Defaults to 1000.
        num_unit_steps (int): The number of the unit steps for annealing. Defaults to 10.
        num_outputs (int): The number of outputs for spin arrays and energy values. Assumed 1 if no value is set. If set to 0, all spin arrays and energy values are output. Defaults to 1.
        filter_solution (bool): Boolean value that specifies whether the solutions are filtered based on given constraints. Default to `False`.
        penalty_calibration (bool): Boolean value that determines whether to automatically adjust the coefficients of the penalty functions of constraint objects. Defaults to `False`.
    """

    amplify_timeout: int = 1000
    num_unit_steps: int = 10
    num_outputs: int = 1
    filter_solution: bool = False
    penalty_calibration: bool = False


class JijFixstarsAmplifySampler(JijModelingInterface):
    jijmodeling_solver_type: SolverType = SolverType(
        queue_name="thirdpartysolver", solver="FixstarsAmplify"
    )

    def __init__(
        self,
        token: str = None,
        url: Union[str, dict] = None,
        proxy=None,
        config=None,
        config_env="default",
        fixstars_token: str = "",
        fixstars_url: str = "",
    ) -> None:
        """Sets Jijzept token and url and fixstars amplify token and url.

        Args:
            token (str, optional): Token string.
            url (Union[str, dict], optional): API URL.
            proxy (str, optional) Proxy URL.
            config (str, optional): Config file path for JijZept.
            config_env (str, optional): config env.
            fixstars_token (str, optional): Token string for Fixstars Amplify.
            fixstars_url (str, optional): Url string for Fixstars Ampplify.
        """
        self.client = JijZeptClient(url, token, proxy, config, config_env)
        self.fixstars_token = fixstars_token
        self.fixstars_url = fixstars_url

    def sample_model(
        self,
        model: Problem,
        feed_dict: dict[str, Union[Number, list, np.ndarray]],
        multipliers: dict[str, Number] = {},
        fixed_variables: dict[str, dict[tuple[int, ...], Number]] = {},
        parameters: Optional[JijFixstarsAmplifyParameters] = None,
        timeout: Optional[int] = None,
        sync=True,
        queue_name: Optional[str] = None,
        **kwargs,
    ) -> DimodResponse:
        """Converts the given problem to amplify.BinaryQuadraticModel and run Fixstars Amplify Solver.
        Note here that the supported type of decision variables is only Binary when using Fixstars Ampplify Solver from Jijzept.

        Args:
            model (Problem): Mathematical expression of JijModeling.
            feed_dict (dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (dict[str, Number]): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (dict[str, dict[tuple[int, ...], Number]], optional): Dictionary of variables to fix.
            parameters (Optional[JijFixstarsAmplifyParameters], optional): Parameters used in Fixstars Amplify. If `None`, the default value of the JijFixstarsAmplifyParameters will be set.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (Optional[str], optional): Queue name.

        Returns:
            dimod.Sampleset: Stores samples and other information.

        Examples:

        ```python
        import jijmodeling as jm
        from jijzept import JijFixstarsAmplifySampler, JijFixstarsAmplifyParameters


        w = jm.Placeholder("w", dim=1)
        num_items = jm.Placeholder("num_items")
        c = jm.Placeholder("c")
        y = jm.Binary("y", shape=(num_items,))
        x = jm.Binary("x", shape=(num_items, num_items))
        i = jm.Element("i", num_items)
        j = jm.Element("j", num_items)
        problem = jm.Problem("bin_packing")
        problem += y[:]
        problem += jm.Constraint("onehot_constraint", jm.Sum(j, x[i, j]) - 1 == 0, forall=i)
        problem += jm.Constraint("knapsack_constraint", jm.Sum(i, w[i] * x[i, j]) - y[j] * c <= 0, forall=j)

        feed_dict = {"num_items": 2, "w": [9, 1], "c": 10}
        multipliers = {"onehot_constraint": 11, "knapsack_constraint": 22}

        sampler = JijFixstarsAmplifySampler(config="xx", amplify_token="oo")
        parameters = JijFixstarsAmplifyParameters(amplify_timeout=1000, num_unit_steps=10, num_outputs=1, filter_solution=False, penalty_calibration=False)
        sampleset = sampler.sample_model(
            problem, feed_dict, multipliers, parameters=parameters
        )
        ```
        """
        if parameters is None:
            parameters = JijFixstarsAmplifyParameters()

        parameters = asdict(parameters)
        if kwargs:
            logging.warning("Setting Parameters using **kwargs is not recommended.")
            parameters.update(**kwargs)

        parameters["timeout"] = parameters.pop("amplify_timeout")
        return super().sample_model(
            model=model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=False,
            num_search=15,
            algorithm=None,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
            token=self.fixstars_token,
            url=self.fixstars_url,
            parameters=parameters,
        )



import logging, sys, time

from logging import getLogger
from typing import Any, Dict, Optional, Type, TypeVar

from jijzept.client import JijZeptClient
from jijzept.response import APIStatus, BaseResponse

ResponseType = TypeVar("ResponseType", bound=BaseResponse)

logger = getLogger(__name__)
handler = logging.StreamHandler(sys.stdout)
logger.addHandler(handler)
logger.setLevel(logging.INFO)

# set the default timeout to 60 min if not specified.
DEFAULT_TIMEOUT_SEC = 3600


def post_instance_and_query(
    ResponseType: Type[TypeVar("ResponseType", bound="BaseResponse")],
    client: JijZeptClient,
    instance_type: str,
    instance: Dict[str, Any],
    queue_name: str,
    solver: str,
    parameters: dict,
    timeout: Optional[float],
    sync: bool = True,
):

    # Instance を投げる
    logger.info("uploading instance ...")
    instance_id = client.post_instance(instance_type, instance)

    # Solverにqueryをなげる
    logger.info("submitting query ...")
    actual_timeout = DEFAULT_TIMEOUT_SEC if timeout is None else timeout
    solver_res = client.submit_solve_query(
        queue_name, solver, parameters, instance_id, actual_timeout
    )
    logger.info(f"submitted to the queue.")
    logger.info(f'Your solution_id is {solver_res["solution_id"]}.')

    # 同期モードで解を取得
    status = "PENDING"
    show_running_flag = False
    if sync:
        polling_count = 0
        while (status == APIStatus.PENDING.value) or (
            status == APIStatus.RUNNING.value
        ):
            response = client.fetch_result(solver_res["solution_id"])
            status = response["status"]
            if (
                response["status"] == APIStatus.RUNNING.value
                and show_running_flag == False
            ):
                logger.info("running...")
                show_running_flag = True
            elif response["status"] == APIStatus.SUCCESS.value:
                break
            elif response["status"] == APIStatus.FAILED.value:
                return ResponseType.empty_response(
                    APIStatus.FAILED,
                    client,
                    solver_res["solution_id"],
                    err_dict=response["solution"],
                )

            elif response["status"] == APIStatus.UNKNOWNERROR.value:
                return ResponseType.empty_response(
                    APIStatus.UNKNOWNERROR, client, solver_res["solution_id"]
                )

            time.sleep(2)
            polling_count = polling_count + 1

            if polling_count == 10:
                # show hints if polling is repeated certain times
                logger.info(f"It takes a lot of time to get the solution...")
                logger.info(f"You can also access the solution by")
                logger.info(f">>> a = XXXSampler(config=...)")
                logger.info(f'>>> a.get_result("{solver_res["solution_id"]}")')

    else:
        logger.info(f"You can access the solution by")
        logger.info(f">>> a = XXXSampler(config=...)")
        logger.info(f'>>> a.get_result("{solver_res["solution_id"]}")')
        return ResponseType.empty_response(
            APIStatus.PENDING, client, solver_res["solution_id"]
        )

    res_obj = ResponseType.from_json_obj(response["solution"])
    res_obj.set_status(APIStatus.SUCCESS)
    return res_obj



from numbers import Number
from typing import Dict, List, Optional, Tuple, Union

import cimod
import numpy as np
import openjij as oj

from jijmodeling.expression.expression import Expression

from jijzept.client import JijZeptClient
from jijzept.entity.schema import SolverType
from jijzept.post_api import post_instance_and_query
from jijzept.response import DimodResponse
from jijzept.sampler.jijmodel_post import JijModelingInterface


class JijSXAuroraSampler(JijModelingInterface):
    jijmodeling_solver_type = SolverType(queue_name="sxaurorasolver", solver="SAParaSearch")

    def __init__(
        self,
        token: str = None,
        url: Union[str, Dict] = None,
        proxy=None,
        config=None,
        config_env="default",
    ):
        """Sets token and url.

        Args:
            token (str, optional): Token string. Defaults to None.
            url (Union[str, Dict], optional): API URL. Defaults to None.
            proxy (str, optional) Proxy URL. Defaults to None.
            config (str, optional): Config file path. Defaults to None.

        Raises:
            :obj:`TypeError`: `token`, `url`, or `config` is not str.
        """
        self.client = JijZeptClient(url, token, proxy, config, config_env)

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_reads: Optional[int] = None,
        tsallis_q: Optional[float] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using JijModeling by means of Microsoft QIO Simulated Annealing solver.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            tsallis_q (Optional[float], optional): q value of tsallis statistics. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import jijmodeling as jm
        n = jm.Placeholder('n')
        x = jm.Binary('x', shape=n)
        d = jm.Placeholder('d', shape=n)
        i = jm.Element("i", n)
        problem = jm.Problem('problem')
        problem += jm.Sum({i: n}, d[i] * x[i])
        sampler = jz.JijSXAuroraSampler(config='config.toml')
        response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
        ```
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            num_sweeps=num_sweeps,
            beta_min=beta_min,
            beta_max=beta_max,
            num_reads=num_reads,
            tsallis_q=tsallis_q,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )




import pathlib

import toml


def load_config(file_name: str, config="default") -> dict:
    """Loads config file (TOML file).

    Args:
        file_name (str): Path to config file.
        config (str, optional): Loading enviroment name. Defaults to 'default'.

    Raises:
        TypeError: If 'config' enviroment is not defined in config file.

    Returns:
        dict: {'token': 'xxxx', 'url': 'xxxx'}
    """
    p_rel = pathlib.Path(file_name)
    with open(p_rel) as f:
        toml_setting_file = toml.load(f)
    if config not in toml_setting_file:
        raise TypeError(
            "'{}' is not defined in config file ({}).".format(config, file_name)
        )
    return toml_setting_file



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

import jijzept.client
import jijzept.client as client
import jijzept.config
import jijzept.config as config
import jijzept.post_api
import jijzept.post_api as post_api
import jijzept.response
import jijzept.response as response
import jijzept.sampler
import jijzept.sampler as sampler
import jijzept.setting
import jijzept.setting as setting

from jijzept.sampler import (
    JijAmazonBraketDWaveSampler,
    JijQIOPASampler,
    JijQIOPTSampler,
    JijQIOSASampler,
    JijQIOSQASampler,
    JijQIOSSMCSampler,
    JijQIOTBSampler,
    JijSASampler,
    JijSQASampler,
    JijSwapMovingSampler,
    JijFixstarsAmplifySampler,
    JijFixstarsAmplifyParameters,
    JijSXAuroraSampler,
    JijDA3SolverParameters,
    JijDA3Sampler,
)

__all__ = [
    "JijAmazonBraketDWaveSampler",
    "JijQIOPASampler",
    "JijQIOPTSampler",
    "JijQIOSASampler",
    "JijQIOSQASampler",
    "JijQIOSSMCSampler",
    "JijQIOTBSampler",
    "JijSASampler",
    "JijSQASampler",
    "JijSwapMovingSampler",
    "JijFixstarsAmplifySampler",
    "JijSXAuroraSampler",
    "JijDA3Sampler",
    "JijDA3SolverParameters",
]



from numbers import Number
from typing import Dict, Optional, Tuple, Union

import cimod
import numpy as np
import openjij

from jijmodeling.expression.expression import Expression

from jijzept.entity.schema import SolverType
from jijzept.response import DimodResponse
from jijzept.sampler.jijmodel_post import JijModelingInterface
from jijzept.sampler.sampler import JijZeptSampler


class JijSASampler(JijZeptSampler, JijModelingInterface):

    solver_type = SolverType(queue_name="openjijsolver", solver="SA")
    hubo_solver_type = SolverType(queue_name="openjijsolver", solver="HUBOSA")
    jijmodeling_solver_type = SolverType(
        queue_name="openjijsolver", solver="SAParaSearch"
    )

    def sample(
        self,
        bqm: Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel],
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        initial_state: Optional[Union[list, dict]] = None,
        updater: Optional[str] = None,
        sparse: Optional[bool] = None,
        reinitialize_state: Optional[bool] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using BinaryQuadraticModel by means of the simulated annealing.

        Args:
            bqm (Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel]): Binary quadratic model.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1 will be set.
            initial_state (Optional[dict], optional): Initial state. If `None`, this will be set automatically.
            updater (Optional[str], optional): Updater algorithm. "single spin flip" or "swendsen wang". If `None`, "single spin flip" will be set.
            sparse (Optional[bool], optional): If `True`, only non-zero matrix elements are stored, which will save memory. If `None`, `False` will be set.
            reinitialize_state (Optional[bool], optional): If `True`, reinitialize state for each run. If `None`, `True` will be set.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        For `cimod.BinaryQuadraticModel` case:

        ```python
        import jijzept as jz
        import cimod
        bqm = cimod.BinaryQuadraticModel({0: -1, 1: -1}, {(0, 1): -1, (1, 2): -1}, "SPIN")
        sampler = jz.JijSASampler(config='config.toml')
        response = sampler.sample(bqm)
        ```

        One can also use `sample_ising` and `sample_qubo` methods.

        For Ising case:

        ```python
        import jijzept as jz
        h = {0: -1, 1: -1, 2: 1, 3: 1}
        J = {(0, 1): -1, (3, 4): -1}
        sampler = jz.JijSASampler(config='config.toml')
        response = sampler.sample_ising(h, J)
        ```

        For QUBO case:

        ```python
        import jijzept as jz
        Q = {(0, 0): -1, (1, 1): -1, (2, 2): 1, (0, 1): -1, (1, 2): 1}
        sampler = jz.JijSASampler(config='config.toml')
        response = sampler.sample_qubo(Q)
        ```
        """

        return super().sample(
            bqm,
            beta_min=beta_min,
            beta_max=beta_max,
            num_sweeps=num_sweeps,
            num_reads=num_reads,
            initial_state=initial_state,
            updater=updater,
            sparse=sparse,
            reinitialize_state=reinitialize_state,
            seed=seed,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_hubo(
        self,
        J: dict,
        vartype: str,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        initial_state: Optional[Union[list, dict]] = None,
        updater: Optional[str] = None,
        reinitialize_state: Optional[bool] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using BinaryPolynomialModel by means of the simulated annealing.

        Args:
            J (dict): Polynomial interactions.
            vartype (str): Variable type. "SPIN" or "BINARY".
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1 will be set.
            initial_state (Optional[dict], optional): Initial state. If `None`, this will be set automatically.
            updater (Optional[str], optional): Updater algorithm. "single spin flip" or "k-local". If `None`, "single spin flip" will be set. One can select "k-local" only for `vartype`="BINARY".
            reinitialize_state (Optional[bool], optional): If `True`, reinitialize state for each run. If `None`, `True` will be set.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        For spin variable case:

        ```python
        import jijzept as jz
        sampler = jz.JijSASampler(config='config.toml')
        J = {(0,): -1, (0, 1): -1, (0, 1, 2): 1}
        response = sampler.sample_hubo(J, "SPIN")
        ```

        For binary variable case:

        ```python
        import jijzept as jz
        sampler = jz.JijSASampler(config='config.toml')
        J = {(0,): -1, (0, 1): -1, (0, 1, 2): 1}
        response = sampler.sample_hubo(J, "BINARY")
        ```

        """

        return super().sample_hubo(
            J,
            vartype=vartype,
            beta_min=beta_min,
            beta_max=beta_max,
            num_sweeps=num_sweeps,
            num_reads=num_reads,
            initial_state=initial_state,
            updater=updater,
            reinitialize_state=reinitialize_state,
            seed=seed,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        updater: Optional[str] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using JijModeling by means of the simulated annealing.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1 will be set.
            updater (Optional[str], optional): Updater algorithm. "single spin flip" or "swendsen wang". If `None`, "single spin flip" will be set.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import jijmodeling as jm
        n = jm.Placeholder('n')
        x = jm.Binary('x', shape=n)
        d = jm.Placeholder('d', shape=n)
        i = jm.Element("i", n)
        problem = jm.Problem('problem')
        problem += jm.Sum({i: n}, d[i] * x[i])
        sampler = jz.JijSASampler(config='config.toml')
        response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
        ```
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            beta_min=beta_min,
            beta_max=beta_max,
            num_sweeps=num_sweeps,
            num_reads=num_reads,
            updater=updater,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )



from typing import Any

import dimod

from jijzept.response.base import BaseResponse


class DimodResponse(BaseResponse, dimod.SampleSet):
    @classmethod
    def from_json_obj(cls, json_obj) -> Any:
        return cls.from_serializable(json_obj)

    @classmethod
    def empty_data(cls) -> Any:
        return cls.from_samples([], "BINARY", 0)

    def __repr__(self):
        return BaseResponse.__repr__(self) + "\n" + dimod.SampleSet.__repr__(self)

    def __str__(self):
        return BaseResponse.__str__(self) + "\n" + dimod.SampleSet.__str__(self)



import enum

from abc import ABCMeta, abstractmethod
from typing import Callable, TypeVar

from jijzept.client import JijZeptClient

_FuncT = TypeVar("_FuncT", bound=Callable)
_TBaseResponse = TypeVar("_TBaseResponse", bound="BaseResponse")


class APIStatus(enum.Enum):
    SUCCESS = "SUCCESS"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    UNKNOWNERROR = "UNKNOWNERROR"


class BaseResponse(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def from_json_obj(cls, json_obj) -> None:
        """abstract method for initializing object from JSON data

        Args:
            json_obj: JSON data
        """

    @classmethod
    @abstractmethod
    def empty_data(cls) -> None:
        """abstract method for generating empty data"""

    def _set_config(self, client: JijZeptClient, solution_id: str):
        self._client = client
        self.solution_id = solution_id

    def set_status(self, status: APIStatus):
        self._status = status

    def set_err_dict(self, err_dict: dict):
        self._err_dict = err_dict

    @property
    def status(self):
        if hasattr(self, "_status"):
            return self._status
        else:
            return APIStatus.PENDING

    @property
    def error_message(self):
        if hasattr(self, "_err_dict"):
            return self._err_dict
        else:
            return {}

    @classmethod
    def empty_response(
        cls, status: APIStatus, client: JijZeptClient, solution_id: str, err_dict={}
    ):
        """generate empty_response

        Args:
            status (APIStatus): status
            client (JijZeptClient): client
            solution_id (str): solution_id
            err_dict:
        """
        response: cls = cls.empty_data()
        response._set_config(client, solution_id)
        response.set_status(status)
        response.set_err_dict(err_dict)
        return response

    def get_result(self: _TBaseResponse) -> _TBaseResponse:
        """get result from cloud.

        If status is updated. update self data
        """
        if self.status in {APIStatus.PENDING, APIStatus.RUNNING}:
            response = self._client.fetch_result(self.solution_id)
            if response["status"] == APIStatus.SUCCESS.value:
                temp_obj = self.from_json_obj(response["solution"])
                temp_obj.set_status(APIStatus.SUCCESS)
                # update myself
                self.__dict__.update(temp_obj.__dict__)
                return self

            elif response["status"] == APIStatus.FAILED.value:
                self.set_status(APIStatus.FAILED)
                # store error info
                self.set_err_dict(response["solution"])
                return self

            elif response["status"] == APIStatus.UNKNOWNERROR.value:
                self.set_status(APIStatus.UNKNOWNERROR)
                return self

            else:
                self.set_status(APIStatus(response["status"]))
                return self
        else:
            return self

    def __repr__(self):
        return_str = self.status.__repr__()
        if self.status == APIStatus.FAILED:
            return_str += "\n"
            return_str += str(self.error_message)

        return return_str

    def __str__(self):
        return_str = self.status.__repr__()
        if self.status == APIStatus.FAILED:
            return_str += "\n"
            return_str += str(self.error_message)

        return return_str



import json

from abc import ABCMeta
from numbers import Number
from typing import Dict, Optional, Tuple, TypeVar, Union

import numpy as np

from jijmodeling import Problem
from jijmodeling.expression.serializable import to_serializable
from jijmodeling.utils import serialize_fixed_var

from jijzept.entity.schema import SolverType
from jijzept.post_api import post_instance_and_query
from jijzept.response import BaseResponse, DimodResponse

ResponseType = TypeVar("ResponseType", bound=BaseResponse)


class JijModelingInterface(metaclass=ABCMeta):
    jijmodeling_solver_type: SolverType

    def sample_model(
        self,
        model: Problem,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Union[float, Number]],
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        timeout=None,
        sync=True,
        queue_name: Optional[str] = None,
        **kwargs,
    ):
        """abstract sample_model.

        Args:
            model (Problem): model
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): feed_dict
            multipliers (Dict[str, Union[float, Number]]): multipliers
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): fixed_variables
            search (bool): if True parameter search is enabled.
            num_search (int): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.
            kwargs:
        """

        m_seri = to_serializable(model)

        parameters = kwargs
        parameters["multipliers"] = multipliers
        parameters["mul_search"] = search
        # fixed variables must be serialized
        parameters["fixed_variables"] = serialize_fixed_var(fixed_variables)
        parameters["num_search"] = num_search
        parameters["algorithm"] = algorithm

        # convert nd.array to list
        _feed_dict = {
            k: v.tolist() if isinstance(v, np.ndarray) else v
            for k, v in feed_dict.items()
        }

        response = post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="JijModeling",
            instance={
                "mathematical_model": json.dumps(m_seri),
                "instance_data": _feed_dict,
            },
            queue_name=self.jijmodeling_solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.jijmodeling_solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

        return response



from numbers import Number
from typing import Dict, List, Optional, Tuple, Union

import cimod
import numpy as np
import openjij as oj

from jijmodeling.expression.expression import Expression

from jijzept.client import JijZeptClient
from jijzept.entity.schema import SolverType
from jijzept.post_api import post_instance_and_query
from jijzept.response import DimodResponse
from jijzept.sampler.jijmodel_post import JijModelingInterface


class JijQIOSASampler(JijModelingInterface):
    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="SA")
    jijmodeling_solver_type = SolverType(queue_name="qiosolver", solver="SAParaSearch")

    def __init__(
        self,
        token: str = None,
        url: Union[str, Dict] = None,
        proxy=None,
        config=None,
        config_env="default",
    ):
        """Sets token and url.

        Args:
            token (str, optional): Token string. Defaults to None.
            url (Union[str, Dict], optional): API URL. Defaults to None.
            proxy (str, optional) Proxy URL. Defaults to None.
            config (str, optional): Config file path. Defaults to None.

        Raises:
            :obj:`TypeError`: `token`, `url`, or `config` is not str.
        """
        self.client = JijZeptClient(url, token, proxy, config, config_env)

    def sample(
        self,
        model: Union[
            cimod.BinaryQuadraticModel,
            oj.BinaryQuadraticModel,
            cimod.BinaryPolynomialModel,
            oj.BinaryPolynomialModel,
        ],
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_reads: Optional[int] = None,
        seed: Optional[int] = None,
        platform: Optional[str] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using BinaryQuadraticModel or BinaryPolynomialModel by means of Microsoft QIO Simulated Annealing solver.

        Args:
            model (Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel, cimod.BinaryPolynomialModel, openjij.BinaryPolynomialModel]): Binary quadratic model or binary polynomial model.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            platform (Optional[str]): "CPU" or "FPGA". If `None`, "CPU" will be set.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import openjij as oj
        bqm = oj.BinaryQuadraticModel({"a": -1}, {("a", "b"):1}, "SPIN")
        bpm = oj.BinaryPolynomialModel({("a", "b", "c", "d"):1}, "BINARY")
        response_bqm = jz.JijQIOSASampler(config='config.toml').sample(bqm_oj)
        response_bpm = jz.JijQIOSASampler(config='config.toml').sample(bpm_oj)
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "num_reads": num_reads,
            "seed": seed,
            "platform": platform,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=model.to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_reads: Optional[int] = None,
        seed: Optional[int] = None,
        platform: Optional[str] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using JijModeling by means of Microsoft QIO Simulated Annealing solver.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            platform (Optional[str]): "CPU" or "FPGA". If `None`, "CPU" will be set.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import jijmodeling as jm
        n = jm.Placeholder('n')
        x = jm.Binary('x', shape=n)
        d = jm.Placeholder('d', shape=n)
        i = jm.Element("i", n)
        problem = jm.Problem('problem')
        problem += jm.Sum({i: n}, d[i] * x[i])
        sampler = jz.JijQIOSASampler(config='config.toml')
        response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
        ```
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            num_sweeps=num_sweeps,
            beta_min=beta_min,
            beta_max=beta_max,
            num_reads=num_reads,
            seed=seed,
            platform=platform,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_ising(
        self,
        h: Dict,
        J: Dict,
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_reads: Optional[int] = None,
        seed: Optional[int] = None,
        platform: Optional[str] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO simulated annealing solver is performed for Ising models.

        Args:
            h (Dict): The linear terms.
            J (Dict): The quadratic terms.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            platform (Optional[str]): "CPU" or "FPGA". If `None`, "CPU" will be set.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOSASampler(config='config.toml')
        response = sampler.sample_ising(h={0: +0.5, 2: -0.5},
                                        J={(0,1):-1, (0,3):-1},)
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "num_reads": num_reads,
            "seed": seed,
            "platform": platform,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryQuadraticModel(h, J, "SPIN").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_qubo(
        self,
        Q: Dict,
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_reads: Optional[int] = None,
        seed: Optional[int] = None,
        platform: Optional[str] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO simulated annealing solver is performed for binary quadratic models.

        Args:
            Q (Dict): The linear or quadratic terms.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            platform (Optional[str]): "CPU" or "FPGA". If `None`, "CPU" will be set.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOSASampler(config='config.toml')
        response = sampler.sample_qubo(Q={(0,0): -0.5, (0,1): -1, (0,2): -1, (2,3): -1})
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "num_reads": num_reads,
            "seed": seed,
            "platform": platform,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(Q, "BINARY").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_hubo(
        self,
        J: Dict,
        vartype: str,
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        num_reads: Optional[int] = None,
        seed: Optional[int] = None,
        platform: Optional[str] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO simulated annealing solver is performed for binary polynomial models.

        Args:
            J (Dict): Polynomial interactions.
            vartype (str): The variable type. "SPIN" or "BINARY".
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            platform (Optional[str]): "CPU" or "FPGA". If `None`, "CPU" will be set.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOSASampler(config='config.toml')
        response = sampler.sample_hubo(J={(0,1,2,3,4): -1}, vartype="SPIN")
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "num_reads": num_reads,
            "seed": seed,
            "platform": platform,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(J, vartype).to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )


class JijQIOPASampler(JijModelingInterface):
    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="PA")
    jijmodeling_solver_type = SolverType(queue_name="qiosolver", solver="PAParaSearch")

    def __init__(
        self,
        token: str = None,
        url: Union[str, Dict] = None,
        proxy=None,
        config=None,
        config_env="default",
    ):
        """Sets token and url.

        Args:
            token (str, optional): Token string. Defaults to None.
            url (Union[str, Dict], optional): API URL. Defaults to None.
            proxy (str, optional) Proxy URL. Defaults to None.
            config (str, optional): Config file path. Defaults to None.

        Raises:
            :obj:`TypeError`: `token`, `url`, or `config` is not str.
        """
        self.client = JijZeptClient(url, token, proxy, config, config_env)

    def sample(
        self,
        model: Union[
            cimod.BinaryQuadraticModel,
            oj.BinaryQuadraticModel,
            cimod.BinaryPolynomialModel,
            oj.BinaryPolynomialModel,
        ],
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        population: Optional[int] = None,
        schedule_type: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using BinaryQuadraticModel or BinaryPolynomialModel by means of Microsoft QIO Population Annealing solver.

        Args:
            model (Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel, cimod.BinaryPolynomialModel, openjij.BinaryPolynomialModel]): Binary quadratic model or binary polynomial model.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            population (Optional[int], optional): The number of walkers in the population (must be positive). If `None`, this will be set automatically.
            schedule_type (Optional[str], optional): "linear" or "geometric".
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import openjij as oj
        bqm = oj.BinaryQuadraticModel({"a": -1}, {("a", "b"):1}, "SPIN")
        bpm = oj.BinaryPolynomialModel({("a", "b", "c", "d"):1}, "BINARY")
        response_bqm = jz.JijQIOPASampler(config='config.toml').sample(bqm_oj)
        response_bpm = jz.JijQIOPASampler(config='config.toml').sample(bpm_oj)
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "population": population,
            "schedule_type": schedule_type,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=model.to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        population: Optional[int] = None,
        schedule_type: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using JijModeling by means of Microsoft QIO Population Annealing solver.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            population (Optional[int], optional): The number of walkers in the population (must be positive). If `None`, this will be set automatically.
            schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import jijmodeling as jm
        n = jm.Placeholder('n')
        x = jm.Binary('x', shape=n)
        d = jm.Placeholder('d', shape=n)
        i = jm.Element("i", n)
        problem = jm.Problem('problem')
        problem += jm.Sum({i: n}, d[i] * x[i])
        sampler = jz.JijQIOPASampler(config='config.toml')
        response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
        ```
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            num_sweeps=num_sweeps,
            beta_min=beta_min,
            beta_max=beta_max,
            population=population,
            schedule_type=schedule_type,
            seed=seed,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_ising(
        self,
        h: Dict,
        J: Dict,
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        population: Optional[int] = None,
        schedule_type: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO population annealing solver is performed for Ising models.

        Args:
            h (Dict): The linear terms.
            J (Dict): The quadratic terms.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            population (Optional[int], optional): The number of walkers in the population (must be positive). If `None`, this will be set automatically.
            schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOPASampler(config='config.toml')
        response = sampler.sample_ising(h={0: +0.5, 2: -0.5},
                                        J={(0,1):-1, (0,3):-1},)
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "population": population,
            "schedule_type": schedule_type,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryQuadraticModel(h, J, "SPIN").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_qubo(
        self,
        Q: Dict,
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        population: Optional[int] = None,
        schedule_type: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO population annealing solver is performed for binary quadratic models.

        Args:
            Q (Dict): The linear or quadratic terms.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            population (Optional[int], optional): The number of walkers in the population (must be positive). If `None`, this will be set automatically.
            schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOPASampler(config='config.toml')
        response = sampler.sample_qubo(Q={(0,0): -0.5, (0,1): -1, (0,2): -1, (2,3): -1})
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "population": population,
            "schedule_type": schedule_type,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(Q, "BINARY").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_hubo(
        self,
        J: Dict,
        vartype: str,
        num_sweeps: Optional[int] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        population: Optional[int] = None,
        schedule_type: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO population annealing solver is performed for binary polynomial models.

        Args:
            J (Dict): Polynomial interactions.
            vartype (str): The variable type. "SPIN" or "BINARY".
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            population (Optional[int], optional): The number of walkers in the population (must be positive). If `None`, this will be set automatically.
            schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOPASampler(config='config.toml')
        response = sampler.sample_hubo(J={(0,1,2,3,4): -1}, vartype="SPIN")
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "population": population,
            "schedule_type": schedule_type,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(J, vartype).to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )


class JijQIOPTSampler(JijModelingInterface):
    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="PT")
    jijmodeling_solver_type = SolverType(queue_name="qiosolver", solver="PTParaSearch")

    def __init__(
        self,
        token: str = None,
        url: Union[str, Dict] = None,
        proxy=None,
        config=None,
        config_env="default",
    ):
        """Sets token and url.

        Args:
            token (str, optional): Token string. Defaults to None.
            url (Union[str, Dict], optional): API URL. Defaults to None.
            proxy (str, optional) Proxy URL. Defaults to None.
            config (str, optional): Config file path. Defaults to None.

        Raises:
            :obj:`TypeError`: `token`, `url`, or `config` is not str.
        """
        self.client = JijZeptClient(url, token, proxy, config, config_env)

    def sample(
        self,
        model: Union[
            cimod.BinaryQuadraticModel,
            oj.BinaryQuadraticModel,
            cimod.BinaryPolynomialModel,
            oj.BinaryPolynomialModel,
        ],
        num_sweeps: Optional[int] = None,
        replicas: Optional[int] = None,
        all_betas: Optional[List[Union[int, float]]] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using BinaryQuadraticModel or BinaryPolynomialModel by means of Microsoft QIO Parallel Tempering solver.

        Args:
            model (Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel, cimod.BinaryPolynomialModel, openjij.BinaryPolynomialModel]): Binary quadratic model or binary polynomial model.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            replicas (Optional[int], optional): The number of concurrent running copies for sampling. Each instance will start with a random configuration.
            all_betas (Optional[List[Union[int, float]]]): The list of beta values used in each replica for sampling. The number of beta values must equal the number of replicas, as each replica will be assigned one beta value from the list.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import openjij as oj
        bqm = oj.BinaryQuadraticModel({"a": -1}, {("a", "b"):1}, "SPIN")
        bpm = oj.BinaryPolynomialModel({("a", "b", "c", "d"):1}, "BINARY")
        response_bqm = jz.JijQIOPTSampler(config='config.toml').sample(bqm_oj)
        response_bpm = jz.JijQIOPTSampler(config='config.toml').sample(bpm_oj)
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "replicas": replicas,
            "all_betas": all_betas,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=model.to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        num_sweeps: Optional[int] = None,
        replicas: Optional[int] = None,
        all_betas: Optional[List[Union[int, float]]] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using JijModeling by means of Microsoft QIO Parallel Tempering solver.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            replicas (Optional[int], optional): The number of concurrent running copies for sampling. Each instance will start with a random configuration.
            all_betas (Optional[List[Union[int, float]]]): The list of beta values used in each replica for sampling. The number of beta values must equal the number of replicas, as each replica will be assigned one beta value from the list.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import jijmodeling as jm
        n = jm.Placeholder('n')
        x = jm.Binary('x', shape=n)
        d = jm.Placeholder('d', shape=n)
        i = jm.Element("i", n)
        problem = jm.Problem('problem')
        problem += jm.Sum({i: n}, d[i] * x[i])
        sampler = jz.JijQIOPTSampler(config='config.toml')
        response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
        ```
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            num_sweeps=num_sweeps,
            replicas=replicas,
            all_betas=all_betas,
            seed=seed,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_ising(
        self,
        h: Dict,
        J: Dict,
        num_sweeps: Optional[int] = None,
        replicas: Optional[int] = None,
        all_betas: Optional[List[Union[int, float]]] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO parallel tempering solver is performed for Ising models.

        Args:
            h (Dict): The linear terms.
            J (Dict): The quadratic terms.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            replicas (Optional[int], optional): The number of concurrent running copies for sampling. Each instance will start with a random configuration.
            all_betas (Optional[List[Union[int, float]]]): The list of beta values used in each replica for sampling. The number of beta values must equal the number of replicas, as each replica will be assigned one beta value from the list.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOPTSampler(config='config.toml')
        response = sampler.sample_ising(h={0: +0.5, 2: -0.5},
                                        J={(0,1):-1, (0,3):-1},)
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "replicas": replicas,
            "all_betas": all_betas,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryQuadraticModel(h, J, "SPIN").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_qubo(
        self,
        Q: Dict,
        num_sweeps: Optional[int] = None,
        replicas: Optional[int] = None,
        all_betas: Optional[List[Union[int, float]]] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO parallel tempering solver is performed for binary quadratic models.

        Args:
            Q (Dict): The linear or quadratic terms.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            replicas (Optional[int], optional): The number of concurrent running copies for sampling. Each instance will start with a random configuration.
            all_betas (Optional[List[Union[int, float]]]): The list of beta values used in each replica for sampling. The number of beta values must equal the number of replicas, as each replica will be assigned one beta value from the list.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOPTSampler(config='config.toml')
        response = sampler.sample_qubo(Q={(0,0): -0.5, (0,1): -1, (0,2): -1, (2,3): -1})
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "replicas": replicas,
            "all_betas": all_betas,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(Q, "BINARY").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_hubo(
        self,
        J: Dict,
        vartype: str,
        num_sweeps: Optional[int] = None,
        replicas: Optional[int] = None,
        all_betas: Optional[List[Union[int, float]]] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO parallel tempering solver is performed for binary polynomial models.

        Args:
            J (Dict): Polynomial interactions.
            vartype (str): The variable type. "SPIN" or "BINARY".
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, this will be set automatically.
            replicas (Optional[int], optional): The number of concurrent running copies for sampling. Each instance will start with a random configuration.
            all_betas (Optional[List[Union[int, float]]]): The list of beta values used in each replica for sampling. The number of beta values must equal the number of replicas, as each replica will be assigned one beta value from the list.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOPTSampler(config='config.toml')
        response = sampler.sample_hubo(J={(0,1,2,3,4): -1}, vartype="SPIN")
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "replicas": replicas,
            "all_betas": all_betas,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(J, vartype).to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )


class JijQIOTBSampler(JijModelingInterface):
    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="TB")
    jijmodeling_solver_type = SolverType(queue_name="qiosolver", solver="TBParaSearch")

    def __init__(
        self,
        token: str = None,
        url: Union[str, Dict] = None,
        proxy=None,
        config=None,
        config_env="default",
    ):
        """Sets token and url.

        Args:
            token (str, optional): Token string. Defaults to None.
            url (Union[str, Dict], optional): API URL. Defaults to None.
            proxy (str, optional) Proxy URL. Defaults to None.
            config (str, optional): Config file path. Defaults to None.

        Raises:
            :obj:`TypeError`: `token`, `url`, or `config` is not str.
        """
        self.client = JijZeptClient(url, token, proxy, config, config_env)

    def sample(
        self,
        model: Union[
            cimod.BinaryQuadraticModel,
            oj.BinaryQuadraticModel,
            cimod.BinaryPolynomialModel,
            oj.BinaryPolynomialModel,
        ],
        num_sweeps: Optional[int] = None,
        tabu_tenure: Optional[int] = None,
        num_reads: Optional[int] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using BinaryQuadraticModel or BinaryPolynomialModel by means of Microsoft QIO Tabu Search solver.

        Args:
            model (Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel, cimod.BinaryPolynomialModel, openjij.BinaryPolynomialModel]): Binary quadratic model or binary polynomial model.
            num_sweeps (Optional[int], optional): The number of steps. If `None`, this will be set automatically.
            tabu_tenure (Optional[int], optional): Tenure of the tabu list in moves. Describes how many moves a variable stays on the tabu list once it has been made.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import openjij as oj
        bqm = oj.BinaryQuadraticModel({"a": -1}, {("a", "b"):1}, "SPIN")
        bpm = oj.BinaryPolynomialModel({("a", "b", "c", "d"):1}, "BINARY")
        response_bqm = jz.JijQIOTBSampler(config='config.toml').sample(bqm_oj)
        response_bpm = jz.JijQIOTBSampler(config='config.toml').sample(bpm_oj)
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "tabu_tenure": tabu_tenure,
            "num_reads": num_reads,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=model.to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        num_sweeps: Optional[int] = None,
        tabu_tenure: Optional[int] = None,
        num_reads: Optional[int] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using JijModeling by means of Microsoft QIO Tabu Search solver.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            num_sweeps (Optional[int], optional): The number of steps. If `None`, this will be set automatically.
            tabu_tenure (Optional[int], optional): Tenure of the tabu list in moves. Describes how many moves a variable stays on the tabu list once it has been made.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import jijmodeling as jm
        n = jm.Placeholder('n')
        x = jm.Binary('x', shape=n)
        d = jm.Placeholder('d', shape=n)
        i = jm.Element("i", n)
        problem = jm.Problem('problem')
        problem += jm.Sum({i: n}, d[i] * x[i])
        sampler = jz.JijQIOTBSampler(config='config.toml')
        response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
        ```
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            num_sweeps=num_sweeps,
            tabu_tenure=tabu_tenure,
            num_reads=num_reads,
            seed=seed,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_ising(
        self,
        h: Dict,
        J: Dict,
        num_sweeps: Optional[int] = None,
        tabu_tenure: Optional[int] = None,
        num_reads: Optional[int] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO tabu search solver is performed for Ising models.

        Args:
            h (Dict): The linear terms.
            J (Dict): The quadratic terms.
            num_sweeps (Optional[int], optional): The number of steps. If `None`, this will be set automatically.
            tabu_tenure (Optional[int], optional): Tenure of the tabu list in moves. Describes how many moves a variable stays on the tabu list once it has been made.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOTBSampler(config='config.toml')
        response = sampler.sample_ising(h={0: +0.5, 2: -0.5},
                                        J={(0,1):-1, (0,3):-1},)
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "tabu_tenure": tabu_tenure,
            "num_reads": num_reads,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryQuadraticModel(h, J, "SPIN").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_qubo(
        self,
        Q: Dict,
        num_sweeps: Optional[int] = None,
        tabu_tenure: Optional[int] = None,
        num_reads: Optional[int] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO tabu search solver is performed for binary quadratic models.

        Args:
            Q (Dict): The linear or quadratic terms.
            num_sweeps (Optional[int], optional): The number of steps. If `None`, this will be set automatically.
            tabu_tenure (Optional[int], optional): Tenure of the tabu list in moves. Describes how many moves a variable stays on the tabu list once it has been made.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOTBSampler(config='config.toml')
        response = sampler.sample_qubo(Q={(0,0): -0.5, (0,1): -1, (0,2): -1, (2,3): -1})
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "tabu_tenure": tabu_tenure,
            "num_reads": num_reads,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(Q, "BINARY").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_hubo(
        self,
        J: Dict,
        vartype: str,
        num_sweeps: Optional[int] = None,
        tabu_tenure: Optional[int] = None,
        num_reads: Optional[int] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO tabu search solver is performed for binary polynomial models.

        Args:
            J (Dict): Polynomial interactions.
            vartype (str): The variable type. "SPIN" or "BINARY".
            num_sweeps (Optional[int], optional): The number of steps. If `None`, this will be set automatically.
            tabu_tenure (Optional[int], optional): Tenure of the tabu list in moves. Describes how many moves a variable stays on the tabu list once it has been made.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOTBSampler(config='config.toml')
        response = sampler.sample_hubo(J={(0,1,2,3,4): -1}, vartype="SPIN")
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "tabu_tenure": tabu_tenure,
            "num_reads": num_reads,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(J, vartype).to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )


class JijQIOSQASampler(JijModelingInterface):
    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="SQA")
    jijmodeling_solver_type = SolverType(queue_name="qiosolver", solver="SQAParaSearch")

    def __init__(
        self,
        token: str = None,
        url: Union[str, Dict] = None,
        proxy=None,
        config=None,
        config_env="default",
    ):
        """Sets token and url.

        Args:
            token (str, optional): Token string. Defaults to None.
            url (Union[str, Dict], optional): API URL. Defaults to None.
            proxy (str, optional) Proxy URL. Defaults to None.
            config (str, optional): Config file path. Defaults to None.

        Raises:
            :obj:`TypeError`: `token`, `url`, or `config` is not str.
        """
        self.client = JijZeptClient(url, token, proxy, config, config_env)

    def sample(
        self,
        model: Union[
            cimod.BinaryQuadraticModel,
            oj.BinaryQuadraticModel,
            cimod.BinaryPolynomialModel,
            oj.BinaryPolynomialModel,
        ],
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        beta: Optional[float] = None,
        trotter: Optional[int] = None,
        transverse_field_max: Optional[float] = None,
        transverse_field_min: Optional[float] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using BinaryQuadraticModel or BinaryPolynomialModel by means of Microsoft QIO Quantum Monte Carlo solver.

        Args:
            model (Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel, cimod.BinaryPolynomialModel, openjij.BinaryPolynomialModel]): Binary quadratic model or binary polynomial model.
            num_sweeps (Optional[int], optional): The number of steps. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            beta (Optional[float], optional): Temperature at which the annealing schedule is executed.
            trotter (Optional[int], optional): The number of copies of every variable to generate for running the simulation.
            transverse_field_max (Optional[float], optional): Starting value of the external field applied to the annealing schedule.
            transverse_field_min (Optional[float], optional): Stopping value of the external field applied to the annealing schedule.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import openjij as oj
        bqm = oj.BinaryQuadraticModel({"a": -1}, {("a", "b"):1}, "SPIN")
        bpm = oj.BinaryPolynomialModel({("a", "b", "c", "d"):1}, "BINARY")
        response_bqm = jz.JijQIOSQASampler(config='config.toml').sample(bqm_oj)
        response_bpm = jz.JijQIOSQASampler(config='config.toml').sample(bpm_oj)
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "num_reads": num_reads,
            "beta": beta,
            "trotter": trotter,
            "transverse_field_max": transverse_field_max,
            "transverse_field_min": transverse_field_min,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=model.to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        beta: Optional[float] = None,
        trotter: Optional[int] = None,
        transverse_field_max: Optional[float] = None,
        transverse_field_min: Optional[float] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using JijModeling by means of Microsoft QIO Quantum Monte Carlo solver.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            num_sweeps (Optional[int], optional): The number of steps. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            beta (Optional[float], optional): Temperature at which the annealing schedule is executed.
            trotter (Optional[int], optional): The number of copies of every variable to generate for running the simulation.
            transverse_field_max (Optional[float], optional): Starting value of the external field applied to the annealing schedule.
            transverse_field_min (Optional[float], optional): Stopping value of the external field applied to the annealing schedule.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import jijmodeling as jm
        n = jm.Placeholder('n')
        x = jm.Binary('x', shape=n)
        d = jm.Placeholder('d', shape=n)
        i = jm.Element("i", n)
        problem = jm.Problem('problem')
        problem += jm.Sum({i: n}, d[i] * x[i])
        sampler = jz.JijQIOSQASampler(config='config.toml')
        response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
        ```
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            num_sweeps=num_sweeps,
            num_reads=num_reads,
            beta=beta,
            trotter=trotter,
            transverse_field_max=transverse_field_max,
            transverse_field_min=transverse_field_min,
            seed=seed,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_ising(
        self,
        h: Dict,
        J: Dict,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        beta: Optional[float] = None,
        trotter: Optional[int] = None,
        transverse_field_max: Optional[float] = None,
        transverse_field_min: Optional[float] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO quantum monte carlo solver is performed for Ising models.

        Args:
            h (Dict): The linear terms.
            J (Dict): The quadratic terms.
            num_sweeps (Optional[int], optional): The number of steps. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            beta (Optional[float], optional): Temperature at which the annealing schedule is executed.
            trotter (Optional[int], optional): The number of copies of every variable to generate for running the simulation.
            transverse_field_max (Optional[float], optional): Starting value of the external field applied to the annealing schedule.
            transverse_field_min (Optional[float], optional): Stopping value of the external field applied to the annealing schedule.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOSQASampler(config='config.toml')
        response = sampler.sample_ising(h={0: +0.5, 2: -0.5},
                                        J={(0,1):-1, (0,3):-1},)
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "num_reads": num_reads,
            "beta": beta,
            "trotter": trotter,
            "transverse_field_max": transverse_field_max,
            "transverse_field_min": transverse_field_min,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryQuadraticModel(h, J, "SPIN").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_qubo(
        self,
        Q: Dict,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        beta: Optional[float] = None,
        trotter: Optional[int] = None,
        transverse_field_max: Optional[float] = None,
        transverse_field_min: Optional[float] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO quantum monte carlo solver is performed for binary quadratic models.

        Args:
            Q (Dict): The linear or quadratic terms.
            num_sweeps (Optional[int], optional): The number of steps. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            beta (Optional[float], optional): Temperature at which the annealing schedule is executed.
            trotter (Optional[int], optional): The number of copies of every variable to generate for running the simulation.
            transverse_field_max (Optional[float], optional): Starting value of the external field applied to the annealing schedule.
            transverse_field_min (Optional[float], optional): Stopping value of the external field applied to the annealing schedule.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOSQASampler(config='config.toml')
        response = sampler.sample_qubo(Q={(0,0): -0.5, (0,1): -1, (0,2): -1, (2,3): -1})
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "num_reads": num_reads,
            "beta": beta,
            "trotter": trotter,
            "transverse_field_max": transverse_field_max,
            "transverse_field_min": transverse_field_min,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(Q, "BINARY").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_hubo(
        self,
        J: Dict,
        vartype: str,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        beta: Optional[float] = None,
        trotter: Optional[int] = None,
        transverse_field_max: Optional[float] = None,
        transverse_field_min: Optional[float] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO quantum monte carlo solver is performed for binary polynomial models.

        Args:
            J (Dict): Polynomial interactions.
            vartype (str): The variable type. "SPIN" or "BINARY".
            num_sweeps (Optional[int], optional): The number of steps. If `None`, this will be set automatically.
            num_reads (Optional[int], optional): The number of samples. If `None`, this will be set automatically.
            beta (Optional[float], optional): Temperature at which the annealing schedule is executed.
            trotter (Optional[int], optional): The number of copies of every variable to generate for running the simulation.
            transverse_field_max (Optional[float], optional): Starting value of the external field applied to the annealing schedule.
            transverse_field_min (Optional[float], optional): Stopping value of the external field applied to the annealing schedule.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOSQASampler(config='config.toml')
        response = sampler.sample_hubo(J={(0,1,2,3,4): -1}, vartype="SPIN")
        ```
        """

        parameters = {
            "num_sweeps": num_sweeps,
            "num_reads": num_reads,
            "beta": beta,
            "trotter": trotter,
            "transverse_field_max": transverse_field_max,
            "transverse_field_min": transverse_field_min,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(J, vartype).to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )


class JijQIOSSMCSampler(JijModelingInterface):
    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="SSMC")
    jijmodeling_solver_type = SolverType(
        queue_name="qiosolver", solver="SSMCParaSearch"
    )

    def __init__(
        self,
        token: str = None,
        url: Union[str, Dict] = None,
        proxy=None,
        config=None,
        config_env="default",
    ):
        """Sets token and url.

        Args:
            token (str, optional): Token string. Defaults to None.
            url (Union[str, Dict], optional): API URL. Defaults to None.
            proxy (str, optional) Proxy URL. Defaults to None.
            config (str, optional): Config file path. Defaults to None.

        Raises:
            :obj:`TypeError`: `token`, `url`, or `config` is not str.
        """
        self.client = JijZeptClient(url, token, proxy, config, config_env)

    def sample(
        self,
        model: Union[
            cimod.BinaryQuadraticModel,
            oj.BinaryQuadraticModel,
            cimod.BinaryPolynomialModel,
            oj.BinaryPolynomialModel,
        ],
        step_limit: Optional[int] = None,
        target_population: Optional[int] = None,
        alpha_min: Optional[float] = None,
        alpha_max: Optional[float] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        alpha_schedule_type: Optional[str] = None,
        beta_schedule_type: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using BinaryQuadraticModel or BinaryPolynomialModel by means of Microsoft QIO Substochastic Monte Carlo solver.

        Args:
            model (Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel, cimod.BinaryPolynomialModel, openjij.BinaryPolynomialModel]): Binary quadratic model or binary polynomial model.
            step_limit (Optional[int], optional): Number of monte carlo steps.
            target_population (Optional[int], optional): The number of walkers in the population (should be greater-equal 8)
            alpha_min (Optional[float], optional): The minimum (final) stepping chance. If `None`, this will be set automatically.
            alpha_max (Optional[float], optional): The maximum (initial) stepping chance. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            alpha_schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            beta_schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import openjij as oj
        bqm = oj.BinaryQuadraticModel({"a": -1}, {("a", "b"):1}, "SPIN")
        bpm = oj.BinaryPolynomialModel({("a", "b", "c", "d"):1}, "BINARY")
        response_bqm = jz.JijQIOSSMCSampler(config='config.toml').sample(bqm_oj)
        response_bpm = jz.JijQIOSSMCSampler(config='config.toml').sample(bpm_oj)
        ```
        """

        parameters = {
            "step_limit": step_limit,
            "target_population": target_population,
            "alpha_min": alpha_min,
            "alpha_max": alpha_max,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "alpha_schedule_type": alpha_schedule_type,
            "beta_schedule_type": beta_schedule_type,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=model.to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        step_limit: Optional[int] = None,
        target_population: Optional[int] = None,
        alpha_min: Optional[float] = None,
        alpha_max: Optional[float] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        alpha_schedule_type: Optional[str] = None,
        beta_schedule_type: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using JijModeling by means of Microsoft QIO Substochastic Monte Carlo solver.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            step_limit (Optional[int], optional): Number of monte carlo steps.
            target_population (Optional[int], optional): The number of walkers in the population (should be greater-equal 8)
            alpha_min (Optional[float], optional): The minimum (final) stepping chance. If `None`, this will be set automatically.
            alpha_max (Optional[float], optional): The maximum (initial) stepping chance. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            alpha_schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            beta_schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import jijmodeling as jm
        n = jm.Placeholder('n')
        x = jm.Binary('x', shape=n)
        d = jm.Placeholder('d', shape=n)
        i = jm.Element("i", n)
        problem = jm.Problem('problem')
        problem += jm.Sum({i: n}, d[i] * x[i])
        sampler = jz.JijQIOSSMCSampler(config='config.toml')
        response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
        ```
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            step_limit=step_limit,
            target_population=target_population,
            alpha_min=alpha_min,
            alpha_max=alpha_max,
            beta_min=beta_min,
            beta_max=beta_max,
            alpha_schedule_type=alpha_schedule_type,
            beta_schedule_type=beta_schedule_type,
            seed=seed,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_ising(
        self,
        h: Dict,
        J: Dict,
        step_limit: Optional[int] = None,
        target_population: Optional[int] = None,
        alpha_min: Optional[float] = None,
        alpha_max: Optional[float] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        alpha_schedule_type: Optional[str] = None,
        beta_schedule_type: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO substochastic monte carlo solver is performed for Ising models.

        Args:
            h (Dict): The linear terms.
            J (Dict): The quadratic terms.
            step_limit (Optional[int], optional): Number of monte carlo steps.
            target_population (Optional[int], optional): The number of walkers in the population (should be greater-equal 8)
            alpha_min (Optional[float], optional): The minimum (final) stepping chance. If `None`, this will be set automatically.
            alpha_max (Optional[float], optional): The maximum (initial) stepping chance. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            alpha_schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            beta_schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOSSMCSampler(config='config.toml')
        response = sampler.sample_ising(h={0: +0.5, 2: -0.5},
                                        J={(0,1):-1, (0,3):-1},)
        ```
        """

        parameters = {
            "step_limit": step_limit,
            "target_population": target_population,
            "alpha_min": alpha_min,
            "alpha_max": alpha_max,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "alpha_schedule_type": alpha_schedule_type,
            "beta_schedule_type": beta_schedule_type,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryQuadraticModel(h, J, "SPIN").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_qubo(
        self,
        Q: Dict,
        step_limit: Optional[int] = None,
        target_population: Optional[int] = None,
        alpha_min: Optional[float] = None,
        alpha_max: Optional[float] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        alpha_schedule_type: Optional[str] = None,
        beta_schedule_type: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO substochastic monte carlo solver is performed for binary quadratic models.

        Args:
            Q (Dict): The linear or quadratic terms.
            step_limit (Optional[int], optional): Number of monte carlo steps.
            target_population (Optional[int], optional): The number of walkers in the population (should be greater-equal 8)
            alpha_min (Optional[float], optional): The minimum (final) stepping chance. If `None`, this will be set automatically.
            alpha_max (Optional[float], optional): The maximum (initial) stepping chance. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            alpha_schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            beta_schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOSSMCSampler(config='config.toml')
        response = sampler.sample_qubo(Q={(0,0): -0.5, (0,1): -1, (0,2): -1, (2,3): -1})
        ```
        """

        parameters = {
            "step_limit": step_limit,
            "target_population": target_population,
            "alpha_min": alpha_min,
            "alpha_max": alpha_max,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "alpha_schedule_type": alpha_schedule_type,
            "beta_schedule_type": beta_schedule_type,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(Q, "BINARY").to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )

    def sample_hubo(
        self,
        J: Dict,
        vartype: str,
        step_limit: Optional[int] = None,
        target_population: Optional[int] = None,
        alpha_min: Optional[float] = None,
        alpha_max: Optional[float] = None,
        beta_min: Optional[float] = None,
        beta_max: Optional[float] = None,
        alpha_schedule_type: Optional[str] = None,
        beta_schedule_type: Optional[str] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Microsoft QIO substochastic monte carlo solver is performed for binary polynomial models.

        Args:
            J (Dict): Polynomial interactions.
            vartype (str): The variable type. "SPIN" or "BINARY".
            step_limit (Optional[int], optional): Number of monte carlo steps.
            target_population (Optional[int], optional): The number of walkers in the population (should be greater-equal 8)
            alpha_min (Optional[float], optional): The minimum (final) stepping chance. If `None`, this will be set automatically.
            alpha_max (Optional[float], optional): The maximum (initial) stepping chance. If `None`, this will be set automatically.
            beta_min (Optional[float], optional): Minimum (initial) inverse temperature. If `None`, this will be set automatically.
            beta_max (Optional[float], optional): Maximum (final) inverse temperature. If `None`, this will be set automatically.
            alpha_schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            beta_schedule_type (Optional[str], optional): "linear" or "geometric". If `None`, "linear" will be set.
            seed (Optional[int], optional): Seed for reproducing results. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        sampler = jz.JijQIOSSMCSampler(config='config.toml')
        response = sampler.sample_hubo(J={(0,1,2,3,4): -1}, vartype="SPIN")
        ```
        """

        parameters = {
            "step_limit": step_limit,
            "target_population": target_population,
            "alpha_min": alpha_min,
            "alpha_max": alpha_max,
            "beta_min": beta_min,
            "beta_max": beta_max,
            "alpha_schedule_type": alpha_schedule_type,
            "beta_schedule_type": beta_schedule_type,
            "seed": seed,
        }

        return post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type="BPM",
            instance=oj.BinaryPolynomialModel(J, vartype).to_serializable(),
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )



from numbers import Number
from typing import Dict, Optional, Tuple, Union

import cimod
import numpy as np
import openjij

from jijmodeling.expression.expression import Expression

from jijzept.entity.schema import SolverType
from jijzept.response import DimodResponse
from jijzept.sampler.jijmodel_post import JijModelingInterface
from jijzept.sampler.sampler import JijZeptSampler


class JijSQASampler(JijZeptSampler, JijModelingInterface):

    solver_type = SolverType(queue_name="openjijsolver", solver="SQA")
    jijmodeling_solver_type = SolverType(
        queue_name="openjijsolver", solver="SQAParaSearch"
    )

    def sample(
        self,
        bqm: Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel],
        beta: Optional[float] = None,
        gamma: Optional[float] = None,
        trotter: Optional[int] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        sparse: Optional[bool] = None,
        reinitialize_state: Optional[bool] = None,
        seed: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using BinaryQuadraticModel by means of the simulated quantum annealing.

        Args:
            bqm (Union[cimod.BinaryQuadraticModel, openjij.BinaryQuadraticModel]): Binary quadratic model.
            beta (Optional[float], optional): Inverse temperature. If `None`, 5.0 will be set.
            gamma (Optional[float], optional): Amplitude of quantum fluctuations. If `None`, 1.0 will be set.
            trotter (Optional[int], optional): The number of trotter slices. It should be an even number.  If `None`, 4 will be set.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1 will be set.
            sparse (Optional[bool], optional): If `True`, only non-zero matrix elements are stored, which will save memory. If `None`, `False` will be set.
            reinitialize_state (Optional[bool], optional): If `True`, reinitialize state for each run. If `None`, `True` will be set.
            seed (Optional[int], optional): Seed for Monte Carlo algorithm. If `None`, this will be set automatically.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:
        One can use like `JijSASampler()` as follows.

        For `cimod.BinaryQuadraticModel` case:

        ```python
        import jijzept as jz
        import cimod
        bqm = cimod.BinaryQuadraticModel({0: -1, 1: -1}, {(0, 1): -1, (1, 2): -1}, "SPIN")
        sampler = jz.JijSQASampler(config='config.toml')
        response = sampler.sample(bqm)
        ```

        One can also use `sample_ising` and `sample_qubo` methods.

        For Ising case:

        ```python
        import jijzept as jz
        h = {0: -1, 1: -1, 2: 1, 3: 1}
        J = {(0, 1): -1, (3, 4): -1}
        sampler = jz.JijSQASampler(config='config.toml')
        response = sampler.sample_ising(h, J)
        ```

        For QUBO case:

        ```python
        import jijzept as jz
        Q = {(0, 0): -1, (1, 1): -1, (2, 2): 1, (0, 1): -1, (1, 2): 1}
        sampler = jz.JijSQASampler(config='config.toml')
        response = sampler.sample_qubo(Q)
        ```
        """

        # The number of trotter should be even
        # since c++ implementation
        if trotter is not None and trotter % 2 == 1:
            raise ValueError("trotter number should be an even number.")

        return super().sample(
            bqm,
            beta=beta,
            gamma=gamma,
            trotter=trotter,
            num_sweeps=num_sweeps,
            num_reads=num_reads,
            sparse=sparse,
            reinitialize_state=reinitialize_state,
            seed=seed,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_model(
        self,
        model: Expression,
        feed_dict: Dict[str, Union[Number, list, np.ndarray]],
        multipliers: Dict[str, Number] = {},
        fixed_variables: Dict[str, Dict[Tuple[int, ...], Union[int, float]]] = {},
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        beta: Optional[float] = None,
        gamma: Optional[float] = None,
        trotter: Optional[int] = None,
        num_sweeps: Optional[int] = None,
        num_reads: Optional[int] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
    ) -> DimodResponse:
        """Sample using JijModeling by means of the simulated quantum annealing.

        Args:
            model (Expression): Mathematical expression of JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            beta (Optional[float], optional): Inverse temperature. If `None`, 5.0 will be set.
            gamma (Optional[float], optional): Amplitude of quantum fluctuations. If `None`, 1.0 will be set.
            trotter (Optional[int], optional): The number of trotter slices. It should be an even number.  If `None`, 4 will be set.
            num_sweeps (Optional[int], optional): The number of Monte-Carlo steps. If `None`, 1000 will be set.
            num_reads (Optional[int], optional): The number of samples. If `None`, 1 will be set.
            timeout (Optional[int], optional): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool, optional): Synchronous mode.
            queue_name (str, optional): queue_name.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:

        ```python
        import jijzept as jz
        import jijmodeling as jm
        n = jm.Placeholder('n')
        x = jm.Binary('x', shape=n)
        d = jm.Placeholder('d', shape=n)
        i = jm.Element("i", n)
        problem = jm.Problem('problem')
        problem += jm.Sum({i: n}, d[i] * x[i])
        sampler = jz.JijSQASampler(config='config.toml')
        response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
        ```
        """

        return super().sample_model(
            model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            beta=beta,
            gamma=gamma,
            trotter=trotter,
            num_sweeps=num_sweeps,
            num_reads=num_reads,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
        )

    def sample_hubo(self, *args, **kwargs):
        """
        This function cannot be used from JijSQASampler.
        """
        raise RuntimeError(
            f"sample_hubo cannot be called from {self.__class__.__name__}"
        )



from __future__ import annotations
from dataclasses import dataclass, asdict
from numbers import Number
from typing import Optional, Union
import logging

import numpy as np

from jijzept.response import DimodResponse
from jijmodeling import Problem
from jijzept.client import JijZeptClient
from jijzept.entity.schema import SolverType
from jijzept.sampler.jijmodel_post import JijModelingInterface


@dataclass
class JijDA3SolverParameters:
    """Manage Parameters for using Digital Annealer v3

    Parameters:
        time_limit_sec (int): Set the timeout in seconds in the range 1 ~ 1800.
        target_energy (Optional[float]): Set the target energy value. The calculation is terminated when the minimum energy savings reaches the target energy.
        num_run (int): Set the number of parallel trial iterations in the range 1 ~ 16.
        num_group (int): Set the number of groups of parallel trials in the range 1 ~ 16.
        num_output_solution (int): Set the number of output solutions for each parallel trial group in the range 1 ~ 1024.
        gs_level (int): Set the level of global search. The higher this value, the longer the constraint utilization search will search in the range 0 ~ 100.
        If set the 1way 1hot or 2way 1hot, it is recommended that 0 be set for gs_level.
        gs_cutoff (int): Set the convergence decision level in the constraint utilization search of the solver in the range 0 ~ 1000000. If `0`, convergence judgement is off.
        penalty_auto_mode (int): Set the coefficient adjustment mode for the constaint term.
        If `0`, fixed to the value setin `penlaty_coef`. If `1`, the value set in `penalty_coef` is used as the initial value and adjusted automatically.
        penalty_coef (int): Set the coefficients of the constraint term.
        penalty_inc_rate (int): Set parameters for automatic adjustment of constriant term.
        max_penalty_coef (int): Set the maximum value of the constraint term coefficient in the global search. If no maximum value is specified, set to 0.
        penalty_strength (float): Set the coefficient when the constraint that appears when converting HUBO to QUBO is treated as a penalty term.
    """

    time_limit_sec: int = 10
    target_energy: Optional[float] = None
    num_run: int = 16
    num_group: int = 1
    num_output_solution: int = 5
    gs_level: int = 5
    gs_cutoff: int = 8000
    penalty_auto_mode: int = 1
    penalty_coef: int = 1
    penalty_inc_rate: int = 150
    max_penalty_coef: int = 0
    penalty_strength: float = 5.0


# TODO: add integration test
class JijDA3Sampler(JijModelingInterface):
    jijmodeling_solver_type: SolverType = SolverType(
        # queue_name="testsolver",
        queue_name="thirdpartysolver",
        solver="DigitalAnnealer",
    )

    def __init__(
        self,
        token: Optional[str] = None,
        url: Union[str, dict] = None,
        proxy: Optional[str] = None,
        config: Optional[str] = None,
        config_env: str = "default",
        da3_token: str = "",
        # da3_url: str = "",
        da3_url: str = "https://api.aispf.global.fujitsu.com/da",
        # da3_url: Optional[str] = None,
    ) -> None:
        """Sets Jijzept token and url and Digital Annealer v3 token and url.

        Args:
            token (Optional[str]): Token string.
            url (Optional[Union[str, dict]]): API URL.
            proxy (Optional[str]): Proxy URL.
            config (Optional[str]): Config file path for JijZept.
            config_env (str): config env.
            da3_token (str): Token string for Degital Annealer 3.
            da3_url (Optional[str]): API Url string for Degital Annealer 3.
        """
        self.client = JijZeptClient(url, token, proxy, config, config_env)
        self.da3_token = da3_token
        self.da3_url = da3_url

    def sample_model(
        self,
        model: Problem,
        feed_dict: dict[str, Union[Number, list, np.ndarray]],
        multipliers: dict[str, Number] = {},
        fixed_variables: dict[str, dict[tuple[int, ...], Number]] = {},
        parameters: Optional[JijDA3SolverParameters] = None,
        search: bool = False,
        num_search: int = 15,
        algorithm: Optional[str] = None,
        timeout: Optional[int] = None,
        sync: bool = True,
        queue_name: Optional[str] = None,
        **kwargs,
    ) -> DimodResponse:
        """Sample using JijModeling by means of Digital Annealer v3.

        Args:
            model (Problem): Mathematical expression of JijModeling.
            feed_dict (dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (dict[str, Number]): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (dict[str, dict[tuple[int, ...], Number]]): Dictionary of variables to fix.
            parameters (Optional[JijDA3SolverParameters]): Parameters used in Ditital Annealer3. If `None`, the default value of the JijDA3SolverParameters will be set.
            search (bool): If `True` parameter search is enabled.
            num_search (int): The number of parameter search iteration. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search.
            timeout (Optional[int]): The number of timeout [sec] for post request. If `None`, 3600 (one hour) will be set.
            sync (bool): Synchronous mode.
            queue_name (Optional[str]): Queue name.
            kwargs: Digital Annealer 3 parameters using **kwags. If both `**kwargs` and `parameters` are exist, the value of `**kwargs` takes precedence.

        Returns:
            DimodResponse: Stores minimum energy samples and other information.

        Examples:

            ```python
            import jijmodeling as jm
            from jijzept import JijDA3Sampler, JijDA3SolverParameters

            w = jm.Placeholder("w", dim=1)
            num_items = jm.Placeholder("num_items")
            c = jm.Placeholder("c")
            y = jm.Binary("y", shape=(num_items,))
            x = jm.Binary("x", shape=(num_items, num_items))
            i = jm.Element("i", num_items)
            j = jm.Element("j", num_items)
            problem = jm.Problem("bin_packing")
            problem += y[:]
            problem += jm.Constraint("onehot_constraint", jm.Sum(j, x[i, j]) - 1 == 0, forall=i)
            problem += jm.Constraint(
                "knapsack_constraint", jm.Sum(i, w[i] * x[i, j]) - y[j] * c <= 0, forall=j
            )

            feed_dict = {"num_items": 2, "w": [9, 1], "c": 10}
            multipliers = {"onehot_constraint": 11, "knapsack_constraint": 22}

            sampler = JijDA3Sampler(config="xx", da3_token="oo")
            parameters = JijDA3SolverParameters()

            sampleset = sampler.sample_model(
                problem, feed_dict, multipliers=multipliers, parameters=parameters
            )
            ```
        """

        if parameters is None:
            parameters = JijDA3SolverParameters()

        parameters = asdict(parameters)
        if kwargs:
            logging.warn(
                "Setting Parameters using **kwargs is not recommended. Use JijDA3SolverParameters class"
            )
            parameters.update(**kwargs)

        return super().sample_model(
            model=model,
            feed_dict=feed_dict,
            multipliers=multipliers,
            fixed_variables=fixed_variables,
            search=search,
            num_search=num_search,
            algorithm=algorithm,
            timeout=timeout,
            sync=sync,
            queue_name=queue_name,
            parameters=parameters,
            token=self.da3_token,
            url=self.da3_url,
        )



import click

from jijzept.config.handle_config import (
    CONFIG_PATH,
    DEFAULT_CONFIG_FILE,
    HOST_URL,
    create_config,
)


@click.group()
def main():
    pass


@main.command()
def create():
    click.echo("--- Welcome to JijZept ---")
    config_path = click.prompt(
        "Please enter the configuration path", default=CONFIG_PATH
    )
    host_url = click.prompt("Please enter the your API's URL", default=HOST_URL)
    token = click.prompt("Please eneter the your token (subscription key)")
    config_file_name = create_config(token, host_url, config_path)
    click.echo("Config file path {}".format(config_file_name))
    click.echo("----- Thank you -----")


@main.command()
def show():
    click.echo(
        "JijZept's default configuration path: {}".format(
            CONFIG_PATH + DEFAULT_CONFIG_FILE
        )
    )

    with open(CONFIG_PATH + DEFAULT_CONFIG_FILE, "r") as f:
        config_str = f.read()

    click.echo("-------------------")
    click.echo(config_str)
    click.echo("-------------------")


if __name__ == "__main__":
    main()



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)


# import jijzept.sampler.dwave
import jijzept.sampler.amazonbraket
import jijzept.sampler.amazonbraket as amazonbraket
import jijzept.sampler.jijmodel_post
import jijzept.sampler.jijmodel_post as jijmodel_post
import jijzept.sampler.sa_cpu
import jijzept.sampler.sa_cpu as sa_cpu
import jijzept.sampler.sampler
import jijzept.sampler.sampler as sampler
import jijzept.sampler.sqa_cpu
import jijzept.sampler.sqa_cpu as sqa_cpu

from jijzept.sampler.amazonbraket import JijAmazonBraketDWaveSampler
from jijzept.sampler.ms_qio import (
    JijQIOPASampler,
    JijQIOPTSampler,
    JijQIOSASampler,
    JijQIOSQASampler,
    JijQIOSSMCSampler,
    JijQIOTBSampler,
)
from jijzept.sampler.sa_cpu import JijSASampler
from jijzept.sampler.sampler import JijZeptSampler
from jijzept.sampler.sqa_cpu import JijSQASampler
from jijzept.sampler.swap_moving import JijSwapMovingSampler
from jijzept.sampler.fixstars_amplify import (
    JijFixstarsAmplifySampler,
    JijFixstarsAmplifyParameters,
)
from jijzept.sampler.sxaurorasampler import JijSXAuroraSampler
from jijzept.sampler.digitalannealer import JijDA3Sampler, JijDA3SolverParameters

__all__ = [
    "JijAmazonBraketDWaveSampler",
    "JijQIOPASampler",
    "JijQIOPTSampler",
    "JijQIOSASampler",
    "JijQIOSQASampler",
    "JijQIOSSMCSampler",
    "JijQIOTBSampler",
    "JijSASampler",
    "JijZeptSampler",
    "JijSQASampler",
    "JijSwapMovingSampler",
    "JijFixstarsAmplifySampler",
    "JijSXAuroraSampler",
    "JijDA3Sampler",
    "JijDA3SolverParameters",
]



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

import jijzept.dev_sampler.dwave
import jijzept.dev_sampler.dwave as dwave
import jijzept.dev_sampler.sbm
import jijzept.dev_sampler.sbm as sbm

from jijzept.dev_sampler.dwave import JijDWaveLeapSampler
from jijzept.dev_sampler.sbm import JijSBMSampler

__all__ = [
    "JijDWaveLeapSampler",
    "JijSBMSampler",
]



from typing import Any, TypeVar, Union

import openjij

from cimod import BinaryPolynomialModel

from jijzept.client import JijZeptClient
from jijzept.entity.schema import SolverType
from jijzept.post_api import post_instance_and_query
from jijzept.response import APIStatus, BaseResponse, DimodResponse

ResponseType = TypeVar("ResponseType", bound=BaseResponse)


# We use openjij.sampler.BaseSampler for Ising and QUBO conversion performance.
class JijZeptSampler(openjij.sampler.sampler.BaseSampler):
    """JijZeptSampler

    another Sampler is based on this class

    """

    solver_type: SolverType
    hubo_solver_type: SolverType

    def __init__(
        self,
        token: str = None,
        url: Union[str, dict] = None,
        proxy=None,
        config=None,
        config_env="default",
    ):
        """Sets token and url.

        Args:
            token (str, optional): Token string. Defaults to None.
            url (Union[str, dict], optional): API URL. Defaults to None.
            proxy (str, optional) Proxy URL. Defaults to None.
            config (str, optional): Config file path. Defaults to None.

        Raises:
            :obj:`TypeError`: `token`, `url`, or `config` is not str.
        """

        self.client = JijZeptClient(url, token, proxy, config, config_env)

    def sample(self, bqm, timeout=None, sync=True, queue_name=None, **kwargs) -> Any:
        """abstract sample function

        Args:
            bqm:
            timeout:
            sync:
            queue_name:
            kwargs:

        Returns:
            Any:
        """
        parameters = kwargs

        instance_type = "BQM"
        instance = bqm.to_serializable()

        response = post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type=instance_type,
            instance=instance,
            queue_name=self.solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )
        return response

    def sample_hubo(
        self, polynomial, vartype, timeout=None, sync=True, queue_name=None, **kwargs
    ) -> Any:
        """abstract sample_hubo.

        Args:
            polynomial:
            vartype:
            timeout:
            sync:
            queue_name:
            kwargs:

        Returns:
            Any:
        """
        parameters = kwargs

        instance_type = "BPM"
        instance = BinaryPolynomialModel(polynomial, vartype=vartype).to_serializable()

        response = post_instance_and_query(
            DimodResponse,
            self.client,
            instance_type=instance_type,
            instance=instance,
            queue_name=self.hubo_solver_type.queue_name
            if queue_name is None
            else queue_name,
            solver=self.hubo_solver_type.solver,
            parameters=parameters,
            timeout=timeout,
            sync=sync,
        )
        return response

    def get_result(self, solution_id: str):
        response = DimodResponse.empty_response(
            APIStatus.PENDING, self.client, solution_id
        )

        return response.get_result()

    @property
    def properties(self):
        return dict()

    @property
    def parameters(self):
        return dict()
